# Příklady na dynamické programování
#
# Jan Kybic

import sys
import time
import memoize

# http://interactivepython.org/runestone/static/pythonds/Recursion/DynamicProgramming.html

def vytiskni_platbu(m,h):
  """ m - pocty minci, h - hodnoty """
  for j in range(len(h)):
    if m[j]>0:
       print("%2d*%2dKč" % (m[j],h[j]), end="")
  print("")


# jak zaplatit částku 'x' pomocí co nejmenšího počtu mincí o hodnotách 'h'
# vrátí pole počtů mincí
def zaplat_nejlepe(h,x):
  def zaplat_internal(x):
    # existuje mince s hodnotou 'x'?
    if x in h:
      return [ int(x==v) for v in h ]
    else: 
      nejlepsi=None
      for i in range(len(h)):
        if h[i]<=x:
          m=zaplat_internal(x-h[i])
          m[i]+=1
          if nejlepsi is None or sum(m)<sum(nejlepsi):
            nejlepsi=m.copy()
          # m[i]-=1
      return nejlepsi
  return zaplat_internal(x)  

# verze s memoizací
def zaplat_nejlepe2(h,x):
  def zaplat_internal(x):
    # existuje mince s hodnotou 'x'?
    if x in h:
      return [ int(x==v) for v in h ]
    else: 
      nejlepsi=None
      for i in range(len(h)):
        if h[i]<=x:
          m=zaplat_internal(x-h[i])
          m[i]+=1
          if nejlepsi is None or sum(m)<sum(nejlepsi):
            nejlepsi=m.copy()
          # m[i]-=1  
      return nejlepsi
  zaplat_internal=memoize.memoize(zaplat_internal)
  return zaplat_internal(x)  

# verze s dynamickým programováním
def zaplat_nejlepe3(h,x):
  reseni=[ [] ] * (x+1)      # zde budeme ukládat zatím nejlepší řešení pro zaplacení 'x' Kč 
  reseni[0]=[0] * len(h)
  for y in range(1,x+1): # jak zaplatit y=1..x Kč
    nejlepsi_i=None      # kterou minci je nejlépe použít
    nejlepsi_pocet=None  # kolik mincí bude pak potřeba
    for i in range(len(h)):   # zkus všechny mince
      if h[i]<=y:             # lze ji použít?
        pocet=sum(reseni[y-h[i]])+1 # ano, tolik mincí je potřeba
        if nejlepsi_i is None or pocet<nejlepsi_pocet:
          nejlepsi_i=i
          nejlepsi_pocet=pocet
    if nejlepsi_i is None:
      print("Nevím, jak zaplatit %d" % y)
      return None
    reseni[y]=reseni[y-h[nejlepsi_i]].copy()
    reseni[y][nejlepsi_i]+=1
  return reseni[x]
    
# --------------------------------------------------------------------------------
  
def test_zaplat_nejlepe(x=733168):
  #x=733168
  #x=24500
  #x=35
  h=[50,20,10,5,2,1]
  if x<=35:
    t=time.clock()
    m=zaplat_nejlepe(h,x)
    t=time.clock()-t
    print("zaplat_nejlepe: %4.2fs" % t)
    vytiskni_platbu(m,h)
  if x<=24500:
    t=time.clock()
    m=zaplat_nejlepe2(h,x)
    t=time.clock()-t
    print("zaplat_nejlepe2: %4.2fs" % t)
    vytiskni_platbu(m,h)
  t=time.clock()
  m=zaplat_nejlepe3(h,x)
  t=time.clock()-t
  print("zaplat_nejlepe3: %4.2fs" % t)
  vytiskni_platbu(m,h)

          
