# Příklady na memoizaci
#
# Jan Kybic, 2016

import time
import functools

def is_prime(p):
  """ Je 'n' prvočíslo? """
  if p<2:
    return False
  n=2
  while n*n<=p:
    if p % n == 0:
      return False
    n+=1
  return True  

def memoize(f):
  """ Vrátí funkci, která se chová jako 'f', ale jednou vypočítané hodnoty jsou zapamatovány
  a nejsou počítány znovu """
  memo={}
  def g(x):
    if x not in memo:
      memo[x]=r=f(x)
      return r
    return memo[x]  
  return g

# Změříme, jak to bude trvat dlouho
  
def apply_f_to_n(f,n):
  """ Změří dobu výpočtu f(1) až f(n) """
  t0=time.clock()
  for i in range(n):
    f(i)
  return time.clock()-t0

def try_memoize_primes():
   n=100000
   apply_f_to_n(is_prime,n)
   print("Bez memoizace:       %4.2fs" % apply_f_to_n(is_prime,n))
   g=memoize(is_prime)
   print("S memoizací poprvé:  %4.2fs" % apply_f_to_n(g,n))
   print("S memoizací podruhé: %4.2fs" % apply_f_to_n(g,n))
   h=functools.lru_cache(maxsize=None)(is_prime)
   print("S lru_cache poprvé:  %4.2fs" % apply_f_to_n(h,n))
   print("S lru_cache podruhé: %4.2fs" % apply_f_to_n(h,n))
   
##################################################
   
def fibonacci(n):
  if n <= 2:
    return 1
  else:
    return fibonacci(n-1) + fibonacci(n-2)

def fibonacci2(n):
  if n <= 2:
    return 1
  else:
    return fibonacci3(n-1) + fibonacci3(n-2)

fibonacci3=memoize(fibonacci2)  

def fibonacci4(n):
  p=[ 1 for i in range(n+1) ]
  for i in range(3,n+1):
    p[i]=p[i-1]+p[i-2]
  return p[n]  



def try_fibonacci():
  # n=40
  # print("n=%d" % n)
  # t0=time.clock() ; fibonacci(n) ; t=time.clock()-t0
  # print("fibonacci:    %4.2fs" % t)
  # t0=time.clock() ; fibonacci3(n) ; t=time.clock()-t0
  # print("fibonacci3:   %4.2fs" % t)
  # t0=time.clock() ; fibonacci3(n) ; t=time.clock()-t0
  # print("fibonacci3:   %4.2fs" % t)
  # t0=time.clock() ; fibonacci4(n) ; t=time.clock()-t0
  # print("fibonacci4:   %4.2fs" % t)
  n=40000
  print("n=%d" % n)
  #t0=time.clock() ; fibonacci3(n) ; t=time.clock()-t0
  #print("fibonacci3:   %4.2fs" % t)
  #t0=time.clock() ; fibonacci3(n) ; t=time.clock()-t0
  #print("fibonacci3:   %4.2fs" % t)
  t0=time.clock() ; fibonacci4(n) ; t=time.clock()-t0
  print("fibonacci4:   %4.2fs" % t)
  
